@@ -47,11 +47,16 @@ summary_interval = {}
 summary_interval_temp = {}
 hashtag_changesets = {}
 countries_changesets = {}
+whitelisted_users = []
 
+print("Initializing ....")
 # read the GeoJSON file
 countries_df = gpd.read_file(
     "https://raw.githubusercontent.com/kshitijrajsharma/OSMSG/master/data/countries_un.geojson"
 )
+geofabrik_countries = pd.read_csv(
+    "https://raw.githubusercontent.com/kshitijrajsharma/OSMSG/master/data/countries.csv"
+)
 
 
 def collect_changefile_stats(
@@ -72,6 +77,20 @@ def collect_changefile_stats(
         except Exception as ex:
             # print("WARNING: way  incomplete." % w.id)
             pass
+
+    # set default
+    users.setdefault(
+        user,
+        {
+            "name": uname,
+            "uid": user,
+            "changesets": 0,
+            "nodes": {"create": 0, "modify": 0, "delete": 0},
+            "ways": {"create": 0, "modify": 0, "delete": 0},
+            "relations": {"create": 0, "modify": 0, "delete": 0},
+            "poi": {"create": 0, "modify": 0},  # nodes that has tags
+        },
+    )
     if summary:
         summary_interval.setdefault(
             timestamp,
@@ -86,177 +105,99 @@ def collect_changefile_stats(
             },
         )
 
-        # for changeset count
+    # changeset count
+    users_temp.setdefault(user, {"changesetrs": []})
+    if summary:
         summary_interval_temp.setdefault(timestamp, {"changesets": [], "users": []})
         if changeset not in summary_interval_temp[timestamp]["changesets"]:
             summary_interval_temp[timestamp]["changesets"].append(changeset)
         summary_interval[timestamp]["changesets"] = len(
             summary_interval_temp[timestamp]["changesets"]
         )
-        # for users count
         if user not in summary_interval_temp[timestamp]["users"]:
             summary_interval_temp[timestamp]["users"].append(user)
         summary_interval[timestamp]["users"] = len(
             summary_interval_temp[timestamp]["users"]
         )
-        # for nodes , ways relations count
+    users_temp[user].setdefault("changesets", [])
+    if changeset not in users_temp[user]["changesets"]:
+        users_temp[user]["changesets"].append(changeset)
+    users[user]["changesets"] = len(users_temp[user]["changesets"])
+
+    # hashtags & countries block
+    if hashtags or changeset_meta:
+        users[user].setdefault("countries", [])
+        users[user].setdefault("hashtags", [])
+
+        try:
+            for ch in hashtag_changesets[changeset]["countries"]:
+                if ch not in users[user]["countries"]:
+                    users[user]["countries"].append(ch)
+            for ch in hashtag_changesets[changeset]["hashtags"]:
+                if ch not in users[user]["hashtags"]:
+                    users[user]["hashtags"].append(ch)
+        except Exception as ex:
+            pass
+
+    # osm element count
+    users[user][osm_type][action] += 1
+    if summary:
         summary_interval[timestamp][osm_type][action] += 1
-        # for poi
-        if osm_type == "nodes" and tags and action != "delete":
+
+    # POI block
+    if osm_type == "nodes" and tags and action != "delete":
+        users[user]["poi"][action] += 1
+        if summary:
             summary_interval[timestamp]["poi"][action] += 1
 
-        # for user supplied tags
-        if tags_to_collect and action != "delete" and tags:
-            for tag in tags_to_collect:
-                summary_interval[timestamp].setdefault(tag, {"create": 0, "modify": 0})
-                if tag in tags:
-                    summary_interval[timestamp][tag][action] += 1
-        # for length calculation
-        if length:
-            for t in length:
-                summary_interval[timestamp].setdefault(f"{t}_create_len", 0)
-                if t in tags and action != "modify" and action != "delete":
-                    summary_interval[timestamp][f"{t}_create_len"] += round(len_feature)
-        # for all tags
-        if all_tags:
+    # all tags block
+    if all_tags:
+        users[user].setdefault("tags_create", {})
+        users[user].setdefault("tags_modify", {})
+        if summary:
             summary_interval[timestamp].setdefault("tags_create", {})
             summary_interval[timestamp].setdefault("tags_modify", {})
-            if tags:
-                for key, value in tags:
-                    if action != "delete":
+        if tags:
+            for key, value in tags:
+                if action != "delete":  # we don't need deleted tags
+                    users[user][f"tags_{action}"].setdefault(key, 0)
+                    users[user][f"tags_{action}"][key] += 1
+                    if summary:
                         summary_interval[timestamp][f"tags_{action}"].setdefault(key, 0)
                         summary_interval[timestamp][f"tags_{action}"][key] += 1
 
-    if user in users:
-        if changeset not in users_temp[user]["changesets"]:
-            users_temp[user]["changesets"].append(changeset)
-        users[user]["changesets"] = len(users_temp[user]["changesets"])
+    # for user supplied tags
+    if tags_to_collect and action != "delete" and tags:
+        for tag in tags_to_collect:
+            if summary:
+                summary_interval[timestamp].setdefault(tag, {"create": 0, "modify": 0})
+            users[user].setdefault(tag, {"create": 0, "modify": 0})
+            if tag in tags:
+                if summary:
+                    summary_interval[timestamp][tag][action] += 1
+                users[user][tag][action] += 1
 
-        if hashtags or changeset:
-            try:
-                for ch in hashtag_changesets[changeset]["countries"]:
-                    if ch not in users[user]["countries"]:
-                        users[user]["countries"].append(ch)
-                for ch in hashtag_changesets[changeset]["hashtags"]:
-                    if ch not in users[user]["hashtags"]:
-                        users[user]["hashtags"].append(ch)
-            except:
-                pass
-
-        users[user][osm_type][action] += 1
-        if osm_type == "nodes" and tags:
-            if action != "delete":
-                users[user]["poi"][action] += 1
-        if all_tags:
-            if tags:
-                for key, value in tags:
-                    if action != "delete":  # we don't need deleted tags
-                        if key in users[user][f"tags_{action}"]:
-                            users[user][f"tags_{action}"][key] += 1
-                        else:
-                            users[user][f"tags_{action}"][key] = 1
-
-        if tags_to_collect:
-            if action != "delete":  # we don't need deleted tags
-                for tag in tags_to_collect:
-                    if tag in tags:
-                        users[user][tag][action] += 1
-        if country:
-            if not hashtags:
-                for ch in countries_changesets[changeset]:
-                    if ch not in users[user]["hashtags"]:
-                        users[user]["hashtags"].append(ch)
-        if length:
-            for t in length:
-                users[user].setdefault(f"{t}_create_len", 0)
-                if (
-                    t in tags
-                    and osm_obj_nodes
-                    and action != "modify"
-                    and action != "delete"
-                ):
-                    users[user][f"{t}_create_len"] += round(len_feature)
-
-    else:
-        users[user] = {
-            "name": uname,
-            "uid": user,
-            "changesets": 0,
-            "nodes": {"create": 0, "modify": 0, "delete": 0},
-            "ways": {"create": 0, "modify": 0, "delete": 0},
-            "relations": {"create": 0, "modify": 0, "delete": 0},
-            "poi": {"create": 0, "modify": 0},  # nodes that has tags
-        }
-        if hashtags or changeset:
-            users[user]["countries"] = []
-            users[user]["hashtags"] = []
-
-            try:
-                users[user]["countries"] = hashtag_changesets[changeset]["countries"]
-                users[user]["hashtags"] = hashtag_changesets[changeset]["hashtags"]
-            except:
-                pass
-
-        if tags_to_collect:
-            for tag in tags_to_collect:
-                users[user][tag] = {"create": 0, "modify": 0}
-        users_temp[user] = {"changesets": []}
-        if changeset not in users_temp[user]["changesets"]:
-            users_temp[user]["changesets"].append(changeset)
-        users[user]["changesets"] = len(users_temp[user]["changesets"])
-        users[user][osm_type][action] = 1
-        if all_tags:
-            users[user]["tags_create"] = {}
-            users[user]["tags_modify"] = {}
-            if tags:
-                for tag, value in tags:
-                    users[user][f"tags_{action}"][tag] = 1
-        if tags_to_collect:
-            for tag in tags_to_collect:
-                if tag in tags:
-                    users[user][tag][action] = 1
-        if country:
-            if not hashtags:
-                users[user]["hashtags"] = countries_changesets[changeset]
+    # for length calculation
+    if length:
+        for t in length:
+            users[user].setdefault(f"{t}_create_len", 0)
+            if summary:
+                summary_interval[timestamp].setdefault(f"{t}_create_len", 0)
+            if t in tags and action != "modify" and action != "delete":
+                if summary:
+                    summary_interval[timestamp][f"{t}_create_len"] += round(len_feature)
+                users[user][f"{t}_create_len"] += round(len_feature)
 
 
 def calculate_stats(
     user, uname, changeset, version, tags, osm_type, timestamp, osm_obj_nodes=None
 ):
-    if (hashtags and country) or hashtags:  # intersect with changesets
+    if hashtags:  # intersect with changesets
         if (
-            len(hashtag_changesets) > 0
+            len(hashtag_changesets) > 0 or len(whitelisted_users) > 0
         ):  # make sure there are changesets to intersect if not meaning hashtag changeset not found no need to go for changefiles
-            if country:
-                if (
-                    changeset in hashtag_changesets.keys()
-                    and changeset in countries_changesets.keys()
-                ):
-                    collect_changefile_stats(
-                        user,
-                        uname,
-                        changeset,
-                        version,
-                        tags,
-                        osm_type,
-                        timestamp,
-                        osm_obj_nodes,
-                    )
-            else:
-                if changeset in hashtag_changesets.keys():
-                    collect_changefile_stats(
-                        user,
-                        uname,
-                        changeset,
-                        version,
-                        tags,
-                        osm_type,
-                        timestamp,
-                        osm_obj_nodes,
-                    )
-    elif country:
-        if len(countries_changesets) > 0:
-            if changeset in countries_changesets.keys():
+
+            if changeset in hashtag_changesets.keys() or uname in whitelisted_users:
                 collect_changefile_stats(
                     user,
                     uname,
@@ -267,6 +208,18 @@ def calculate_stats(
                     timestamp,
                     osm_obj_nodes,
                 )
+    elif len(whitelisted_users) > 0:
+        if uname in whitelisted_users:
+            collect_changefile_stats(
+                user,
+                uname,
+                changeset,
+                version,
+                tags,
+                osm_type,
+                timestamp,
+                osm_obj_nodes,
+            )
     else:  # collect everything
         collect_changefile_stats(
             user, uname, changeset, version, tags, osm_type, timestamp, osm_obj_nodes
@@ -279,7 +232,7 @@ class ChangesetHandler(osmium.SimpleHandler):
 
     def changeset(self, c):
         run_hashtag_check_logic = False
-        if changeset and not hashtags:
+        if changeset_meta and not hashtags:
             if "comment" in c.tags:
                 run_hashtag_check_logic = True
         if hashtags:
@@ -317,32 +270,6 @@ class ChangesetHandler(osmium.SimpleHandler):
                         if row["name"] not in hashtag_changesets[c.id]["countries"]:
                             hashtag_changesets[c.id]["countries"].append(row["name"])
 
-        if country:
-            if c.bounds:
-                bounds = str(c.bounds)
-                if "invalid" not in bounds:
-                    bbox_list = bounds.strip("()").split(" ")
-
-                    minx, miny = bbox_list[0].split("/")
-                    maxx, maxy = bbox_list[1].split("/")
-
-                    bbox = box(float(minx), float(miny), float(maxx), float(maxy))
-                    # Create a point for the centroid of the bounding box
-                    centroid = bbox.centroid
-                    intersected_rows = countries_df[countries_df.intersects(centroid)]
-                    for i, row in intersected_rows.iterrows():
-
-                        if row["name"] == country:
-                            if c.id not in countries_changesets.keys():
-                                countries_changesets[c.id] = []
-                            if "comment" in c.tags:
-                                hashtags_comment = re.findall(
-                                    r"#[\w-]+", c.tags["comment"]
-                                )
-                                for tag in hashtags_comment:
-                                    if tag not in countries_changesets[c.id]:
-                                        countries_changesets[c.id].append(tag)
-
 
 class ChangefileHandler(osmium.SimpleHandler):
     def __init__(self):
@@ -353,34 +280,52 @@ class ChangefileHandler(osmium.SimpleHandler):
             version = n.version
             if n.deleted:
                 version = 0
-            calculate_stats(
-                n.uid, n.user, n.changeset, version, n.tags, "nodes", n.timestamp
-            )
+            try:
+                calculate_stats(
+                    n.uid, n.user, n.changeset, version, n.tags, "nodes", n.timestamp
+                )
+            except Exception as ex:
+                print(f"Warning: {n.id} parse error")
+                print(ex)
 
     def way(self, w):
         if w.timestamp >= start_date_utc and w.timestamp < end_date_utc:
             version = w.version
             if w.deleted:
                 version = 0
-            calculate_stats(
-                w.uid,
-                w.user,
-                w.changeset,
-                version,
-                w.tags,
-                "ways",
-                w.timestamp,
-                w.nodes if length else None,
-            )
+            try:
+                calculate_stats(
+                    w.uid,
+                    w.user,
+                    w.changeset,
+                    version,
+                    w.tags,
+                    "ways",
+                    w.timestamp,
+                    w.nodes if length else None,
+                )
+            except Exception as ex:
+                print(f"Warning: {w.id} parse error")
+                print(ex)
 
     def relation(self, r):
         if r.timestamp >= start_date_utc and r.timestamp < end_date_utc:
             version = r.version
             if r.deleted:
                 version = 0
-            calculate_stats(
-                r.uid, r.user, r.changeset, version, r.tags, "relations", r.timestamp
-            )
+            try:
+                calculate_stats(
+                    r.uid,
+                    r.user,
+                    r.changeset,
+                    version,
+                    r.tags,
+                    "relations",
+                    r.timestamp,
+                )
+            except Exception as ex:
+                print(f"Warning: {r.id} parse error")
+                print(ex)
 
 
 def process_changefiles(url):
@@ -451,8 +396,9 @@ def parse_args():
     )
     parser.add_argument(
         "--country",
+        nargs="+",
         default=None,
-        help="Country name to extract (get name from data/un_countries) : Only viable until day stats since changeset replication is available for minute, avoid using for geofabrik url since geofabrik already gives country level changefiles",
+        help="List of country name to extract (get id from data/countries), It will use geofabrik countries updates so it will require OSM USERNAME. Only Available for Daily Updates",
     )
 
     parser.add_argument(
@@ -492,6 +438,14 @@ def parse_args():
         help="No. of top rows to extract , to extract top 100 , pass 100",
     )
 
+    parser.add_argument(
+        "--users",
+        type=str,
+        nargs="+",
+        default=None,
+        help="List of user names to look for , You can use it to only produce stats for listed users or pass it with hashtags , it will act as or filter. Case sensitive use ' ' to enter names with space in between",
+    )
+
     parser.add_argument(
         "--workers",
         type=int,
@@ -563,7 +517,7 @@ def parse_args():
 
     parser.add_argument(
         "--changeset",
-        help="Include hashtag and country informations on the stats. It forces script to process changeset replciation , Careful to use this since changeset replication is minutely",
+        help="Include hashtag and country informations on the stats. It forces script to process changeset replciation , Careful to use this since changeset replication is minutely according to your internet speed and cpu cores",
         action="store_true",
         default=False,
     )
@@ -605,7 +559,6 @@ def parse_args():
 
 
 def main():
-    print("OSMSG : Waking up ......")
     args = parse_args()
     if args.start_date:
         start_date = strip_utc(
@@ -635,14 +588,25 @@ def main():
 
         end_date = strip_utc(end_date, args.timezone)
     if args.country:
-        if not countries_df["name"].isin([args.country]).any():
-            print("Country doesn't exists : Refer to data/countries_un.csv name column")
-            sys.exit()
+        osc_url_temp = []
+        for ctr in args.country:
+            if not geofabrik_countries["id"].isin([ctr.lower()]).any():
+                print(
+                    f"Error : {ctr} doesn't exists : Refer to data/countries.csv id column"
+                )
+                sys.exit()
+            osc_url_temp.append(
+                geofabrik_countries.loc[
+                    geofabrik_countries["id"] == ctr.lower(), "update_url"
+                ].values[0]
+            )
+        print(f"Ignoring --url , and using Geofabrik Update URL for {args.country}")
+        args.url = osc_url_temp
     if args.changeset:
-        if args.hashtags or args.country:
+        if args.hashtags:
             assert (
                 args.changeset
-            ), "You can not use include changeset meta option along with hashtags/country"
+            ), "You can not use include changeset meta option along with hashtags"
 
     start_time = time.time()
 
@@ -651,17 +615,15 @@ def main():
     global all_tags
     global hashtags
     global length
-    global country
-    global changeset
+    global changeset_meta
     global exact_lookup
     global summary
 
     all_tags = args.all_tags
     additional_tags = args.tags
     hashtags = args.hashtags
-    country = args.country
     cookies = None
-    changeset = args.changeset
+    changeset_meta = args.changeset
     exact_lookup = args.exact_lookup
     length = args.length
     summary = args.summary
@@ -686,6 +648,7 @@ def main():
 
         if any("geofabrik" in url.lower() for url in args.url):
             if args.username is None:
+                # print(os.environ.get("OSM_USERNAME"))
                 args.username = os.environ.get("OSM_USERNAME")
             if args.password is None:
                 args.password = os.environ.get("OSM_PASSWORD")
@@ -712,6 +675,10 @@ def main():
         )
         sys.exit()
 
+    if args.users:
+        for u in args.users:
+            whitelisted_users.append(u)
+
     if args.last_hour:
         start_date, end_date = get_prev_hour(args.timezone)
 
@@ -750,17 +717,22 @@ def main():
     if start_date == end_date:
         print("Err: Start date and end date are equal")
         sys.exit()
-
+    if ((end_date - start_date).days < 1 or args.last_hour) and args.country:
+        print(
+            "Warning : Use --changeset Option to include country info on stats , You can filter your country from output itself for stats lesser than a day, Use --country option for more than Day Statistics"
+        )
+        sys.exit()
     if (end_date - start_date).days > 1:
-        if args.hashtags or args.country:
+
+        if args.hashtags:
             print(
-                "Warning : Replication for Changeset is minutely , To process more than a day data it might take a while, Consider supplying geofabrik country replication url for countries. Use --force to ignore this warning"
+                "Warning : Replication for Changeset is minutely , To download more than day data it might take a while depending upon your internet speed, Use --force to ignore this warning"
             )
             if not args.force:
                 sys.exit()
     print(f"Supplied start_date: {start_date} and end_date: {end_date}")
 
-    if args.hashtags or args.country or args.changeset:
+    if args.hashtags or args.changeset:
 
         Changeset = ChangesetToolKit()
         (
@@ -906,26 +878,27 @@ def main():
                                 )
                             )
                         )
-        summary_df = pd.json_normalize(list(summary_interval.values()))
-        summary_df = summary_df.assign(
-            changes=summary_df["nodes.create"]
-            + summary_df["nodes.modify"]
-            + summary_df["nodes.delete"]
-            + summary_df["ways.create"]
-            + summary_df["ways.modify"]
-            + summary_df["ways.delete"]
-            + summary_df["relations.create"]
-            + summary_df["relations.modify"]
-            + summary_df["relations.delete"]
-        )
-        summary_df.insert(3, "map_changes", summary_df["changes"], True)
-        summary_df = summary_df.drop(columns=["changes"])
+        if args.summary:
+            summary_df = pd.json_normalize(list(summary_interval.values()))
+            summary_df = summary_df.assign(
+                changes=summary_df["nodes.create"]
+                + summary_df["nodes.modify"]
+                + summary_df["nodes.delete"]
+                + summary_df["ways.create"]
+                + summary_df["ways.modify"]
+                + summary_df["ways.delete"]
+                + summary_df["relations.create"]
+                + summary_df["relations.modify"]
+                + summary_df["relations.delete"]
+            )
+            summary_df.insert(3, "map_changes", summary_df["changes"], True)
+            summary_df = summary_df.drop(columns=["changes"])
 
-        summary_df = summary_df.sort_values("timestamp", ascending=True)
+            summary_df = summary_df.sort_values("timestamp", ascending=True)
 
         df = pd.json_normalize(list(users.values()))
 
-        if hashtags or changeset:
+        if hashtags or changeset_meta:
             df["countries"] = df["countries"].apply(lambda x: ",".join(map(str, x)))
 
         df = df.assign(
@@ -958,7 +931,7 @@ def main():
             # Reindex the DataFrame with the new order of column names
             df = df.reindex(columns=cols)
 
-        if country or hashtags or changeset:
+        if hashtags or changeset_meta:
             df["hashtags"] = df["hashtags"].apply(lambda x: ",".join(map(str, x)))
             column_to_move = "hashtags"
             df = df.assign(**{column_to_move: df.pop(column_to_move)})
